
//
// Tiny-Diamonds 0.9
//
// Small game similar to the C-64 classic BoulderDash.
// Source-code is in the public domain for study and entertainment.
// No restrictions on the use of the source-code in ANY way.
// Beware: Use code and executable at own risk, no warrenty what so ever!
//
// Keyboard control:
//  Cursor keys     move rockford (main character of the game)
//  CTRL or SHIFT   perform fetch/pickup of diamond or push of boulder without actually moving.
//  ESC             give up (after closed-in by boulders for example)
//  N               Skip this level and continue with next (some will consider this cheating!)
//
//  CTRL or SHIFT at main screen to start new game.
// 
// Bugs:
//   - Rockford animation not complete, he moves a little static.
//   - Timer doesn't work, so levels will play forever if not finished by user (you might consider that a feature!)
//   - Game should have more levels
//
// Reason for building this beasty:
//   First attempt at a 'real' game engine. Nice playground for experiments with
//   game-loops, timing, keyboard entry and algorithmic textures. Most of the code
//   can be reused to build a platform game or something else which maps on a 2d grid.
//
// Have fun,
//
// Peter
// http://www.syntiac.com
//
//
//
// History
//   0.90 -> 0.91
//     - Transformation from Amoeba to boulders (when count exceeds maximum) was not
//       always successfull due to wrong compare expression. Changed handling of MASK
//

#ifndef _DEBUG

#pragma optimize("agsw", on)
#pragma optimize("y", off)
#pragma comment(linker,"/opt:nowin98 /ENTRY:entry /MAP /MAPINFO:FIXUPS")  /* VC6: turn off 4kbyte alignment. */
#endif

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#pragma warning(disable:4201)
#include <mmsystem.h>
#pragma warning(default:4201)


static HINSTANCE myInstance;
static HWND mainWindow;
static ATOM mainClassAtom;
static const char className[]="TDIAMONDS32";
static const char windowTitle[]="Tiny-Diamonds 0.93";

#define MAXX    128
#define MAXY    128
#define STARTUPTIME       64  // Cycles to wait before transforming start -> rockford
#define AMOEBA_SPEED       5  // Amoeba grow speed, higher is faster
#define MAXAMOEBACOUNT   200  // After Amoeba has reached this size it becomes boulders

typedef int RGB; // 32 bits color (24rgb + 8alpha)
static RGB     *bitplanes;
static HBITMAP playfieldBitmap;
static HDC     playfieldDC;



/////////////////////////////////////////////////////////////////////////////
//
// Graphic engine constants
//
/////////////////////////////////////////////////////////////////////////////
#define SCREENX (20*32)
#define SCREENY (12*32)

#define BITMAPX (SCREENX+96) // 32 for scrolling and 2*32 for objects moving at border
#define BITMAPY (SCREENY+96)

static unsigned char rndArray[32][32]; // Precalculated random numbers
                                       // Used for drawing dirt and boulders.

/////////////////////////////////////////////////////////////////////////////
//
// Current game information
//
/////////////////////////////////////////////////////////////////////////////

static unsigned char playArea[MAXY][MAXX];
static unsigned char moveArea[MAXY][MAXX]; // Implements smooth moving in 4 steps.
// s000yyxx   xx=00 no move  xx=01 right  xx=11 left
//            yy=00 no move  yy=01 down   yy=11 up
//            s=scanned flag



//static RGB colors[4]={0x00000000, 0x00805000, 0x00808080, 0x00FFFFFF};

// 0=Backgound  1=DirtColor  2=SteelColor  3=WhiteHighlight
// 4 to 7 copies of 0-4 to allow color modification and restore (only entry 4 is used)
static RGB colors[8];
static int amoebaCount;
static bool amoebaIsAlive;
static int cycle;
static int flash; // Short bright flash when enough diamonds are collected
static int men;   // Total number of RockFords available to player if 0 then game over
static int score; // Score for player
static int diamondsCollected;

// level info
static bool levelQuit;
static int diamondsNeeded;
static int lastLevel; // Active level before rockford died, so we can display this on intro screen.
static int curLevel;  // Current level, 0 is intro-screen which has special handling in the code.
static int levelX; // X-Size of level
static int levelY; // Y-Size of level
static int curLevelOffs;
static int nextLevelOffs;

// keyboard control
static int rockfortNotSeen;
static int idle;
static int forcekeyboardX; // Take atleast one step when keypressed (gives smoother response on keyboard control)
static int forcekeyboardY;
static int keyboardX;
static int keyboardY;
static int keyboardPickup;
static int keyboardEscape;
static int useJoystick;

static int screenX;
static int screenY;
static int rockfordX;
static int rockfordY;


/////////////////////////////////////////////////////////////////////////////
//
// Info line
//
/////////////////////////////////////////////////////////////////////////////
static void clearInfoLine(void) {
  int yc;

  for(yc=0;yc<24;yc++) {
    memset(bitplanes+yc*BITMAPX, 0, sizeof(RGB)*BITMAPX);
  }
}

// 0x0001 0x0002 0x0004
// 0x0008 0x0010 0x0020
// 0x0040 0x0080 0x0100
// 0x0200 0x0400 0x0800
// 0x1000 0x2000 0x4000
//
static void drawChar(int pos, short chr) {
  int x,y,xc,yc;
  for(y=0;y<5;y++) {
    for(x=0;x<3;x++) {
      for(yc=0;yc<3;yc++) {
        for(xc=0;xc<3;xc++) {
          bitplanes[(y*4+yc+2)*BITMAPX+pos*16+x*4+xc]=(chr&1)?0x00FF0000:0x00500000;
        }
      }
      chr>>=1;
    }
  }
}

static void drawStr(int pos, const short *str) {
  while ((*str)>=0) {
    drawChar(pos++, *(str++));
  }
}

// 0x0001 0x0002 0x0004
// 0x0008 0x0010 0x0020
// 0x0040 0x0080 0x0100
// 0x0200 0x0400 0x0800
// 0x1000 0x2000 0x4000
static const short numbersChars[]={
 0x2B6A, 0x4934, 0x752A, 0x38A3, 0x49EC, 0x38CF, 0x7BCE, 0x24A7, 0x2AAA, 0x29AA,
 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
};
static void drawInt(int pos, int value, int cnt, int base) {
  int p;
  for(p=0;p<cnt;p++) {
    if (value==0 && p!=0) {
      // leading zero suppression
      drawChar(pos+cnt-p-1, 0x0000);
    } else {
      drawChar(pos+cnt-p-1, numbersChars[value%base]);
      value/=base;
    }
  }
}

/////////////////////////////////////////////////////////////////////////////
//
// Playfield draw routines
//
/////////////////////////////////////////////////////////////////////////////

// Get random char
static unsigned char getRnd(void) {
  static rnd;
  rnd=rnd*16807+GetTickCount();
  return (rnd>>24)&255;
}

static void fillRndArray(void) {
  int x,y;
  int rnd=16807;

  for(y=0;y<32;y++) {
    for(x=0;x<32;x++) {
      rnd=rnd*16807;
      rndArray[y][x]=(rnd>>24);
    }
  }
}

// Only 4 colors are used in 4 groups packed to 8 bits
// %%aabbccdd  aa - eyes closed   dd - eyes open
static const unsigned char rockfordTopPict[16][16]={
 {0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00},
 {0x00,0x00,0x00,0x00,0x00,0x00,0xAA,0xAA,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00},
 {0x00,0x00,0x00,0x00,0x00,0x00,0xAA,0xAA,0x00,0x00,0x00,0xAA,0xAA,0xAA,0xAA,0xAA},
 {0x00,0x00,0x00,0x00,0x00,0x00,0xAA,0xAA,0x00,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA},
 {0x00,0x00,0x00,0x00,0x00,0x00,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA},
 {0x00,0x00,0x00,0x00,0xAA,0xAA,0xAA,0xAA,0xA8,0xA8,0xA8,0xAA,0xAA,0xAA,0xAA,0xAA},
 {0x00,0x00,0x00,0xAA,0xAA,0xAA,0xAA,0xA8,0xA0,0xA0,0xA0,0xA8,0xAA,0xAA,0xAA,0xAA},
 {0x00,0x00,0xAA,0x55,0xAA,0xAA,0xAA,0xA0,0xA0,0xA0,0xA0,0xA0,0xAA,0xAA,0xAA,0xAA},
 {0x00,0x00,0xAA,0x55,0xAA,0xAA,0xAA,0x80,0x95,0x95,0x80,0x80,0xAA,0xAA,0xAA,0xAA},
 {0x00,0x00,0x00,0xAA,0xAA,0xAA,0xAA,0xAA,0x80,0x80,0x80,0xAA,0xAA,0xAA,0xAA,0xAA},
 {0x00,0x00,0x00,0x00,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA},
 {0x00,0x00,0x00,0x00,0x00,0x00,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA},
 {0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA},
 {0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA},
 {0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xAA,0xAA,0xAA,0xAA},
 {0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xAA,0xAA,0xAA,0xAA},
};
static const unsigned char rockfordBottomPict[16][16]={
 {0x00,0x00,0x00,0x00,0x00,0x00,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA},
 {0x00,0x00,0x00,0x00,0x00,0xAA,0xAA,0xAA,0xAA,0xAA,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF},
 {0x00,0x00,0x00,0x00,0x00,0xAA,0xAA,0x00,0x00,0x00,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF},
 {0x00,0x00,0x00,0x00,0x00,0xAA,0xAA,0x00,0x00,0x00,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA},
 {0x00,0x00,0x00,0x00,0x00,0xAA,0xAA,0x00,0x00,0x00,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA},
 {0x00,0x00,0x00,0x00,0x00,0xAA,0xAA,0x00,0x00,0x00,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF},
 {0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xFF,0xFF,0xFF,0xFF,0xFF,0xFF},
 {0x00,0x00,0x00,0x00,0x00,0x00,0x00,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA},
 {0x00,0x00,0x00,0x00,0x00,0x00,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA},
 {0x00,0x00,0x00,0x00,0x00,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA,0xAA},
 {0x00,0x00,0x00,0x00,0x00,0xAA,0xAA,0xAA,0x00,0x00,0x00,0xFF,0xFF,0xFF,0xFF,0xFF},
 {0x00,0x00,0x00,0x00,0x00,0xAA,0xAA,0xAA,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00},
 {0x00,0x00,0x00,0x00,0x00,0xAA,0xAA,0xAA,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00},
 {0x00,0x00,0x00,0x00,0x00,0xAA,0xAA,0xAA,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00},
 {0x00,0x00,0x00,0x00,0x00,0xAA,0xAA,0xAA,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00},
 {0x00,0x00,0x00,0xFF,0xFF,0xFF,0xFF,0xFF,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00},
};
/*
{
 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,0,0,0,0,0,0,2,2,2,2,2,0,0,0,0,0,0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,0,0,0,2,2,2,2,2,2,2,2,2,2,0,0,0,0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,0,0,2,2,2,2,2,2,2,2,2,2,2,2,2,0,0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,0,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,0,2,2,2,2,2,2,2,2,2,2,2,2,2,0,0,2,2,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,2,2,2,2,2,2,2,2,1,2,2,2,2,0,0,0,0,2,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,2,2,2,2,2,2,2,1,1,2,2,2,2,0,0,0,0,2,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,2,2,2,2,2,2,2,2,1,2,2,2,2,0,0,0,0,2,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,2,2,2,2,2,2,2,2,2,2,2,2,2,2,0,0,2,2,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,0,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,0,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,0,0,2,2,2,2,2,2,2,2,2,2,2,2,2,0,0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,0,0,0,2,2,2,2,2,2,2,2,2,2,0,0,0,0,0,0,0,0,0,0,0,
 0,0,0,0,0,0,0,0,0,0,0,0,2,2,2,2,2,2,2,2,0,0,0,0,0,0,0,0,0,0,0,0,
}
};*/

static void drawSpace(RGB *p) {
  int xc,yc;

  for(yc=0;yc<32;yc++) {
    for(xc=0;xc<32;xc++) {
      *(p++)=colors[0];
    }
    p=p+BITMAPX-32;
  }
}

static const char dirtColors[]={0,1,1,2};
static void drawDirt(RGB *p) {
  int xc,yc;
  //int rnd=1;

  for(yc=0;yc<32;yc++) {
    for(xc=0;xc<32;xc++) {
      *(p++)=colors[dirtColors[rndArray[yc][xc]&3]];
    }
    p=p+BITMAPX-32;
  }
}

static void drawExplosion(RGB *p) {
  int xc,yc;

  for(yc=0;yc<32;yc++) {
    for(xc=0;xc<32;xc++) {
      int c;
      switch(rndArray[yc][xc]&((8<<(cycle&3))-1)) {
      case 1: c=1; break;
      case 2: c=1; break;
      case 3: c=2; break;
      default: c=0; break;
      }
      *(p++)=colors[c];
    }
    p=p+BITMAPX-32;
  }
}

static void drawBoulder(RGB *p) {
  int xc,yc;

  for(yc=0;yc<32;yc++) {
    for(xc=0;xc<32;xc++) {
      int c;
//      if (((xc-16)*(xc-16)+(yc-16)*(yc-16))>240) {
      if (((xc*xc-32*xc+256)+(yc*yc-32*yc+256))>240) {
        c=0;
      } else {
        c=((rndArray[yc][xc]&31)+xc+(64-yc))>>5;
      }
      *(p++)=colors[c];
    }
    p=p+BITMAPX-32;
  }
}

static void drawExplosive(RGB *p) {
  int xc,yc;

  for(yc=0;yc<32;yc++) {
    for(xc=0;xc<32;xc++) {
      int c;
//      if (((xc-16)*(xc-16)+(yc-16)*(yc-16))>240) {
      if (((xc*xc-32*xc+256)+(yc*yc-32*yc+256))>240) {
        c=0;
      } else {
//        if (((xc-16)*(xc-16)+(yc-16)*(yc-16))>180) {
        if (((xc*xc-32*xc+256)+(yc*yc-32*yc+256))>180) {
          c=2;
        } else {
          c=cycle&3;
        }
      }
      *(p++)=colors[c];
    }
    p=p+BITMAPX-32;
  }
}

static void drawBlockMetal(RGB *p) {
  int xc,yc;

  for(yc=0;yc<32;yc++) {
    for(xc=0;xc<32;xc++) {
      int c;
      int d=((xc-15)*(xc-15)+(yc-15)*(yc-15));
      if (d>360) {
        c=0;
      } else if (xc<2 || yc<2 || xc>29 || yc>29 || d>300) {
        c=1;
      } else {
        c=2;
      }
      *(p++)=colors[c];
    }
    p=p+BITMAPX-32;
  }
}

static void drawBrickWall(RGB *p) {
  int xc,yc;

  for(yc=0;yc<32;yc++) {
    for(xc=0;xc<32;xc++) {
      int c=(((xc+(yc&8))&0x0C)==0 || (yc&0x06)==0)?0:3;
      *(p++)=colors[c];
    }
    p=p+BITMAPX-32;
  }
}

static void drawDiamond(RGB *p) {
  int xc,yc;

  for(yc=0;yc<32;yc++) {
    for(xc=0;xc<32;xc++) {
      int c;
      if (18-xc<yc && xc-14<yc && xc+14>yc && -xc+46>yc) {
        c=((cycle+yc)/4)&3;
      } else if (16-xc<yc && xc-16<yc && xc+16>yc && -xc+48>yc) {
        c=(xc>16)?3:1;
      } else {
        c=0;
      }
      *(p++)=colors[c];
    }
    p=p+BITMAPX-32;
  }
}

static void drawSteelWall(RGB *p) {
  int xc,yc;

  for(yc=0;yc<32;yc++) {
    for(xc=0;xc<32;xc++) {
      int c;
      if ((xc&0x0C)==0x04 && (yc&0x0C)==0x08) {
        c=3;
//      } else if (((xc&0x0C)==0x04 || (xc&0x0C)==0x08) && ((yc&0x0C)==0x04 || (yc&0x0C)==0x08)) {
      } else if (((xc-4)&0x08)==0 && ((yc-4)&0x08)==0) {
        c=0;
      } else {
        c=2;
      }
      *(p++)=colors[c];
    }
    p=p+BITMAPX-32;
  }
}

static const unsigned char amoebaTable[]={100,150,200,250,250,200,150,100};
static void drawAmoeba(RGB *p) {
  int xc,yc;

  for(yc=0;yc<32;yc++) {
    for(xc=0;xc<32;xc++) {
      int c;
      int d=((xc-16)*(xc-16)+(yc-16)*(yc-16));
      int r=rndArray[yc][xc]+amoebaTable[cycle&7];

      if (d>r) {
        c=0;
      } else if (d*2>r) {
        c=2;
      } else {
        c=3;
      }
      *(p++)=colors[c];
    }
    p=p+BITMAPX-32;
  }
}

static void drawStart(RGB *p) {
  if ((cycle&0x04)==0) {
    int xc,yc;

    for(yc=0;yc<32;yc++) {
      for(xc=0;xc<32;xc++) {
        *(p++)=colors[(xc<3 || xc>=29 || yc<3 || yc>=29)?2:0];
      }
      p=p+BITMAPX-32;
    }
  } else {
    drawSteelWall(p);
  }
}

static const char eyesequence[]={0,2,4,6,6,4,2,0};
static void drawRockFord(RGB *p) {
  int xc,yc;

  for(yc=0;yc<16;yc++) {
    for(xc=0;xc<16;xc++) {
      int c;
      if (/*idle>32 &&*/ ((cycle/8)&47)==0) {
        c=(rockfordTopPict[yc][xc]>>eyesequence[cycle&7])&3;
      } else {
        c=rockfordTopPict[yc][xc]&3;
      }
      *(p+xc)=colors[c];
      *(p+(31-xc))=colors[c];
      *(p+xc+16*BITMAPX)=colors[rockfordBottomPict[yc][xc]&3];
      *(p+(31-xc)+16*BITMAPX)=colors[rockfordBottomPict[yc][xc]&3];
//      *(p+yc*BITMAPX+xc)=colors[rockfordTopPict[4][yc][xc]];
    }
    p=p+BITMAPX;
  }
}

static const char firecolors[]={1,0,1,3,0,1,0,1};
static void drawFireFly(RGB *p) {
  int xc, yc;
  
  for(yc=0;yc<32;yc++) {
    for(xc=0;xc<32;xc++) {
      int c=(cycle&3)+3;
      int i;
      for(i=0;i<12;i+=4) {
        if (xc<12-i || xc>20+i || yc<12-i || yc>20+i) {
          c--;
        }
      }
      *(p++)=colors[firecolors[c]];
    }
    p=p+BITMAPX-32;
  }
}

static inline int absi(int v) {
  return (v<0)?-v:v;
}
static const char butterflylist[]={1,1,2,3,4,5,6,7,7,6,5,4,3,2,1,1};
static void drawButterFly(RGB *p) {
  int xc, yc;

  for(yc=0;yc<32;yc++) {
    int c=((yc-8)>>2)&3;
    for(xc=0;xc<32;xc++) {
//      *(p++)=colors[((xc<yc/2 && 32-yc>xc) || (xc>yc && xc<32-(cycle&7) && 32-yc<xc))?((yc>>1)&3):0];
      int sx=absi(16-xc)*butterflylist[cycle&15];
      *(p++)=colors[(xc>=butterflylist[cycle&15]*2 && xc<=32-butterflylist[cycle&15]*2 && sx>16-yc && sx>yc-16)?c:0];
    }
    p=p+BITMAPX-32;
  }
}



/////////////////////////////////////////////////////////////////////////////
//
// Object info
//
/////////////////////////////////////////////////////////////////////////////

static void dummyxy(int x, int y) {}
static void calcExplSpace(int x, int y);
static void calcExplDiamond(int x, int y);
static void calcBoulder(int x, int y);
static void calcFallingBoulder(int x, int y);
static void calcDiamond(int x, int y);
static void calcFallingDiamond(int x, int y);
static void pickupDiamond(int x, int y);
static void calcAmoeba(int x, int y);
static void calcStart(int x, int y);
static void calcExit(int x, int y);
static void calcRockford(int x, int y);
static void calcEnterExit(int x, int y);
static void calcButterFireFly(int x, int y);


#define SCANNED 128
#define MASK    (~SCANNED)
#define t true
#define f false
// Object property flags
//#define F_EMPTY               1
#define F_DESTRUCTABLE        2
#define F_ROUNDEDGE           4 // If set boulders/diamonds can roll off object.
#define F_EXPLODE_ON_HIT      8
#define F_CAN_PUSH_X         16 // Can be moved in X direction by pushing (boulders/blocks)
#define F_CAN_PUSH_Y         32 // Can be moved in Y direction by pushing (blocks)
#define F_KILLS_MONSTER      64 // Will kill monsters on touch (Player/Amoeba)


// Keep following defines in sync with 'Objects' table below
#define O_SPACE               0
#define O_DIRT                1
#define O_BOULDER             2
#define O_FALLINGBOULDER      3
#define O_DIAMOND             4
#define O_FALLINGDIAMOND      5
#define O_BRICKWALL           6
#define O_STEELWALL           7
#define O_START               8
#define O_EXIT                9
#define O_EXITOPEN           10
#define O_ROCKFORD           11
#define O_AMOEBA             12
#define O_BLOCK              14
#define O_EXPLOSIVE          15
#define O_FIREFLY            16
#define O_BUTTERFLY          20
#define O_EXPLOSION_SPACE    24
#define O_EXPLOSION_DIAMONDS 25


static const struct _Objects{
  unsigned char flags;
  unsigned char explodeObject;
  void (*drawFunc)(RGB *p);
  void (*doFunc)(int x, int y);
  void (*pickupFunc[2])(int x, int y); // 0 without buttondown  1 with buttondown
} Objects[]={
/*  0 - Space */      { F_DESTRUCTABLE, 255, drawSpace, dummyxy, {dummyxy,dummyxy} },
/*  1 - Dirt  */      { F_DESTRUCTABLE, 255, drawDirt, dummyxy, {dummyxy,dummyxy} },
/*  2 - Boulder */    { F_DESTRUCTABLE | F_ROUNDEDGE | F_CAN_PUSH_X, 255, drawBoulder, calcBoulder, NULL },
/*  3 - F. Boulder */ { F_DESTRUCTABLE, 255, drawBoulder, calcFallingBoulder, NULL },
/*  4 - Diamond */    { F_DESTRUCTABLE | F_ROUNDEDGE, 255, drawDiamond, calcDiamond, {pickupDiamond,pickupDiamond} },
/*  5 - F. Diamond */ { F_DESTRUCTABLE, 255, drawDiamond, calcFallingDiamond, {pickupDiamond,pickupDiamond} },
/*  6 - Brick wall */ { F_DESTRUCTABLE | F_ROUNDEDGE, 255, drawBrickWall, dummyxy, {NULL,NULL} },
/*  7 - Steel wall */ { 0, 255, drawSteelWall, dummyxy, {NULL,NULL} },
/*  8 - Start pos. */ { F_DESTRUCTABLE, 255, drawStart, calcStart, {NULL,NULL} },
/*  9 - Exit close */ { F_DESTRUCTABLE, 255, drawSteelWall, calcExit, {NULL,NULL} },
/* 10 - Exit open */  { F_DESTRUCTABLE, 255, drawStart, dummyxy, {calcEnterExit, NULL} },
/* 11 - Rockford */   { F_DESTRUCTABLE | F_EXPLODE_ON_HIT | F_KILLS_MONSTER, O_EXPLOSION_SPACE, drawRockFord, calcRockford, NULL },
/* 12 - Amoeba */     { F_DESTRUCTABLE | F_KILLS_MONSTER, 255, drawAmoeba, calcAmoeba, {NULL, NULL}},
/* 13 - dummy */      { 0, 255, NULL, NULL, {NULL, NULL}},
/* 14 - Movable Blk */{ F_ROUNDEDGE | F_CAN_PUSH_X | F_CAN_PUSH_Y, 255, drawBlockMetal, dummyxy, NULL },
/* 15 - Explosive */  { F_DESTRUCTABLE | F_ROUNDEDGE | F_EXPLODE_ON_HIT | F_CAN_PUSH_X, O_EXPLOSION_SPACE, drawExplosive, dummyxy, NULL },
/* Keep firefly/butterfly together on 8 number boundary in this order,
   calcButterFireFly depends on that! */
/* 16 - FireFly L */  { F_DESTRUCTABLE | F_EXPLODE_ON_HIT, O_EXPLOSION_SPACE, drawFireFly, calcButterFireFly, NULL },
/* 17 - FireFly D */  { F_DESTRUCTABLE | F_EXPLODE_ON_HIT, O_EXPLOSION_SPACE, drawFireFly, calcButterFireFly, NULL },
/* 18 - FireFly R */  { F_DESTRUCTABLE | F_EXPLODE_ON_HIT, O_EXPLOSION_SPACE, drawFireFly, calcButterFireFly, NULL },
/* 19 - FireFly U */  { F_DESTRUCTABLE | F_EXPLODE_ON_HIT, O_EXPLOSION_SPACE, drawFireFly, calcButterFireFly, NULL },
/* 20 - ButterFly D */{ F_DESTRUCTABLE | F_EXPLODE_ON_HIT, O_EXPLOSION_DIAMONDS, drawButterFly, calcButterFireFly, NULL },
/* 21 - ButterFly L */{ F_DESTRUCTABLE | F_EXPLODE_ON_HIT, O_EXPLOSION_DIAMONDS, drawButterFly, calcButterFireFly, NULL },
/* 22 - ButterFly U */{ F_DESTRUCTABLE | F_EXPLODE_ON_HIT, O_EXPLOSION_DIAMONDS, drawButterFly, calcButterFireFly, NULL },
/* 23 - ButterFly R */{ F_DESTRUCTABLE | F_EXPLODE_ON_HIT,  O_EXPLOSION_DIAMONDS, drawButterFly, calcButterFireFly, NULL },
/* 24 - Expl Space */ { F_DESTRUCTABLE, 255, drawExplosion, calcExplSpace, NULL },
/* 25 - Expl Diam. */ { F_DESTRUCTABLE, 255, drawExplosion, calcExplDiamond, NULL },
/* end of list */     { 0, 255, NULL, NULL, NULL },
};



/////////////////////////////////////////////////////////////////////////////
//
// Draw playfield
//
/////////////////////////////////////////////////////////////////////////////

static void displayPlayArea(void) {
  int x,y;

  for(y=0;y<(BITMAPY/32)-2;y++) {
    for(x=0;x<(BITMAPX/32)-2;x++) {
      int yc=y+(screenY>>5);
      int xc=x+(screenX>>5);
      if (moveArea[yc][xc] || playArea[yc][xc]==O_SPACE) { // Special treatment for space, for smooth movement.
        drawSpace(bitplanes+y*32*BITMAPX+32*BITMAPX+x*32+32);
      }
    }
  }
  for(y=0;y<(BITMAPY/32)-2;y++) {
    for(x=0;x<(BITMAPX/32)-2;x++) {
      int yc=y+(screenY>>5);
      int xc=x+(screenX>>5);
      RGB *p=bitplanes+y*32*BITMAPX+32*BITMAPX+x*32+32;
      switch(moveArea[yc][xc] & MASK) {
      case  1: // right
        p=p+8*(cycle&3)-32;
        break;
      case  3: // left
        p=p+32-8*(cycle&3);
        break;
      case  4: // down
        p=p+BITMAPX*8*(cycle&3)-32*BITMAPX;
        break;
      case 12: // up
        p=p+32*BITMAPX-BITMAPX*8*(cycle&3);
        break; 
      default:
        break;
      }
      if (playArea[yc][xc]!=O_SPACE) { // Special treatment for space, for smooth movement.
        (Objects[(playArea[yc][xc])].drawFunc)(p);
      }
    }
  }
}

/////////////////////////////////////////////////////////////////////////////
//
// Game logic
//
/////////////////////////////////////////////////////////////////////////////

static void calcExplSpace(int x, int y) {
  playArea[y][x]=O_SPACE;
}

static void calcExplDiamond(int x, int y) {
  playArea[y][x]=O_DIAMOND;
}

static void objectExploded(int x, int y, int replaceObject) {
  int xc, yc;
 
  if (replaceObject!=255) {
    for(yc=-1;yc<2;yc++) {
      for(xc=-1;xc<2;xc++) {
        if (Objects[playArea[y+yc][x+xc]].flags&F_DESTRUCTABLE) {
          playArea[y+yc][x+xc]=replaceObject;
          moveArea[y+yc][x+xc]=SCANNED;
        }
      }
    }
  }
}

static void calcBoulderDiamondWillFall(int x, int y, int fallingObject, int notFallingObject) {
  if (playArea[y+1][x]==0) {
    playArea[y+1][x]=fallingObject;
    playArea[y][x]=O_SPACE;
    moveArea[y+1][x]=4 | SCANNED;
  } else if (Objects[playArea[y+1][x]].flags&F_ROUNDEDGE) {
    if ((playArea[y][x-1]==0) && (playArea[y+1][x-1]==0)) {
      playArea[y][x-1]=fallingObject;
      playArea[y][x]=O_SPACE;
      moveArea[y][x-1]=3 | SCANNED;
    } else if ((playArea[y][x+1]==0) && (playArea[y+1][x+1]==0)) {
      playArea[y][x+1]=fallingObject;
      playArea[y][x]=O_SPACE;
      moveArea[y][x+1]=1 | SCANNED;
    } else {
      playArea[y][x]=notFallingObject; // Clear falling flag
    }
  } else {
    playArea[y][x]=notFallingObject; // Clear falling flag
  }
}

static bool calcBoulderDiamondFalling(int x, int y, int fallingObject, int notFallingObject) {
  if (Objects[playArea[y+1][x]].flags&F_EXPLODE_ON_HIT) {
//    playArea[y][x]=notFallingObject; // In case explosion fails
    objectExploded(x, y+1, Objects[playArea[y+1][x]].explodeObject);
    return true;
  }
  return false;
}

static void calcBoulder(int x, int y) {
  calcBoulderDiamondWillFall(x, y, O_FALLINGBOULDER, O_BOULDER);
}

static void calcFallingBoulder(int x, int y) {
  if (!calcBoulderDiamondFalling(x, y, O_FALLINGBOULDER, O_BOULDER)) {
    calcBoulder(x, y);
  }
}

static void calcDiamond(int x, int y) {
  calcBoulderDiamondWillFall(x, y, O_FALLINGDIAMOND, O_DIAMOND);
}

static void calcFallingDiamond(int x, int y) {
  if (!calcBoulderDiamondFalling(x, y, O_FALLINGDIAMOND, O_DIAMOND)) {
    calcDiamond(x, y);
  }
}

static void pickupDiamond(int x, int y) {
  diamondsCollected++;
  score+=(diamondsCollected>diamondsNeeded)?20:10;
  if (diamondsCollected==diamondsNeeded) {    
    colors[0]=colors[3];
    flash=1;
  }
}

static void calcAmoeba(int x, int y) {
  int nx=0;
  int ny=0;
  bool moved=false;

  amoebaCount++;
  if (playArea[y][x+1]<2) {
    amoebaIsAlive=true;
    if (!moved && getRnd()<AMOEBA_SPEED) { nx=1; moved=true; }
  }
  if (playArea[y+1][x]<2) {
    amoebaIsAlive=true;
    if (!moved && getRnd()<AMOEBA_SPEED) { ny=1; moved=true; }
  }
  if (playArea[y][x-1]<2) {
    amoebaIsAlive=true;
    if (!moved && getRnd()<AMOEBA_SPEED) { nx=-1; moved=true; }
  }
  if (playArea[y-1][x]<2) {
    amoebaIsAlive=true;
    if (!moved && getRnd()<AMOEBA_SPEED) { ny=-1; moved=true; }
  }
  if (cycle>STARTUPTIME) {
    playArea[y+ny][x+nx]=O_AMOEBA;
    moveArea[y+ny][x+nx]=SCANNED;
  }
}

static void calcStart(int x, int y) {
  rockfordX=x;
  rockfordY=y;
  rockfortNotSeen=0;
  if (cycle>STARTUPTIME) {
    playArea[y][x]=O_ROCKFORD;
  }
}

static void calcExit(int x, int y) {
  if (diamondsCollected>=diamondsNeeded) {
    playArea[y][x]=O_EXITOPEN;
  }
}

static void calcEnterExit(int x, int y) {
  levelQuit=true;
}

static void calcRockford(int x, int y) {
  bool moved=false;
  int keyx;
  int keyy;
  if (forcekeyboardX || forcekeyboardY) {
    keyx=forcekeyboardX;
    keyy=forcekeyboardY;
  } else {
    keyx=keyboardX;
    keyy=keyboardY;
  }

  rockfortNotSeen=0;
  if (keyboardEscape) {
    objectExploded(x, y, Objects[playArea[y][x]].explodeObject);
    keyboardEscape=0;
  } else if (Objects[playArea[y+keyy][x+keyx]].pickupFunc[keyboardPickup]) {
    (*Objects[playArea[y+keyy][x+keyx]].pickupFunc[keyboardPickup])(x, y);
    playArea[y+keyy][x+keyx]=O_SPACE;
    moved=true;
  } else if (((keyy==0 && (Objects[playArea[y][x+keyx]].flags&F_CAN_PUSH_X)) ||
              (keyx==0 && (Objects[playArea[y+keyy][x]].flags&F_CAN_PUSH_Y))) &&
             ((playArea[y+2*keyy][x+2*keyx])==0)) {
    playArea[y+2*keyy][x+2*keyx]=playArea[y+keyy][x+keyx];
    playArea[y+keyy][x+keyx]=O_SPACE;
    moved=true;
  }
  if (moved && keyboardPickup==0) {
    idle=0;
    rockfordX=x+keyx;
    rockfordY=y+keyy;
    playArea[y+keyy][x+keyx]=O_ROCKFORD;
    playArea[y][x]=O_SPACE;
    moveArea[y+keyy][x+keyx]=(((keyy&3)<<2)+(keyx&3)) | SCANNED;
  } else {
    idle++;
    rockfordX=x;
    rockfordY=y;
  }
}

static const signed char bfX1[]={  0,  1,  0, -1, -1,  0,  1,  0};
static const signed char bfX2[]={ -1,  0,  1,  0,  0, -1,  0,  1};
static const signed char bfY1[]={  1,  0, -1,  0,  0, -1,  0,  1};
static const signed char bfY2[]={  0,  1,  0, -1,  1,  0, -1,  0};
static void calcButterFireFly(int x, int y) {
  int i=playArea[y][x]&7;

  if ((Objects[playArea[y][x+1]].flags |
       Objects[playArea[y][x-1]].flags |
       Objects[playArea[y+1][x]].flags |
       Objects[playArea[y-1][x]].flags) & F_KILLS_MONSTER) {
    objectExploded(x, y, Objects[playArea[y][x]].explodeObject);
  } else if ((playArea[y+bfY1[i]][x+bfX1[i]])==0) {
    /* Turn corner */
    playArea[y+bfY1[i]][x+bfX1[i]]=((playArea[y][x]&(~3))+((i+1)&3));
    playArea[y][x]=O_SPACE;
    moveArea[y+bfY1[i]][x+bfX1[i]]=(((bfY1[i]&3)<<2)+(bfX1[i]&3)) | SCANNED;
  } else if ((playArea[y+bfY2[i]][x+bfX2[i]])==0) {
    /* Keep going in current direction */
    playArea[y+bfY2[i]][x+bfX2[i]]=playArea[y][x];
    playArea[y][x]=O_SPACE;
    moveArea[y+bfY2[i]][x+bfX2[i]]=(((bfY2[i]&3)<<2)+(bfX2[i]&3)) | SCANNED;
  } else {
    /* Turn because we ran against obstacle */
    playArea[y][x]=((playArea[y][x]&(~3))+((i-1)&3));
  }
}

// Replace all amoeba with new object
static void amoebaToObject(int obj) {
  int x,y;

  for(y=0;y<levelY;y++) {
    for(x=0;x<levelX;x++) {
      if (playArea[y][x]==O_AMOEBA) {
        playArea[y][x]=obj;
      }
    }
  }
}

static void calculateArea(void) {
  int x,y;
  
  // house-keeping jobs
  amoebaCount=0;
  amoebaIsAlive=false;
  rockfortNotSeen++;
  if (flash) flash--; else colors[0]=colors[4];

  // Reset scanned/moved flags
  for(y=0;y<levelY;y++) {
    for(x=0;x<levelX;x++) {
      moveArea[y][x]=0;
    }
  }
  // Perform required action for each object.
  for(y=0;y<levelY;y++) {
    for(x=0;x<levelX;x++) {
      if ((moveArea[y][x]&SCANNED)==0) {
        (*Objects[playArea[y][x]].doFunc)(x, y);
      }
    }
  }
  // Amoeba is dead (becomes diamonds)
  if (amoebaIsAlive==false && amoebaCount>0) {
    amoebaToObject(O_EXPLOSION_DIAMONDS);
  }
  // Amoeba is grown to large (becomes boulders)
  if (amoebaCount>MAXAMOEBACOUNT) {
    amoebaToObject(O_BOULDER);
  }

  forcekeyboardX=0;
  forcekeyboardY=0;

/*
      switch() {
      case OBJ_GROWWALL:
        if (playArea[y][x-1]==OBJ_SPACE) {
          playArea[y][x-1]=OBJ_GROWWALL | OBJ_SCANNED;
        }
        if (playArea[y][x+1]==OBJ_SPACE) {
          playArea[y][x+1]=OBJ_GROWWALL | OBJ_SCANNED;
        }
        break;
      }
    }
  }
*/
}

/////////////////////////////////////////////////////////////////////////////
//
// Levels
//
/////////////////////////////////////////////////////////////////////////////

#include "compressedlevels.h"


// Make sure this matches with the table in the level compressor!
static const unsigned char codeToObject[]={
  O_SPACE, O_DIRT, O_BOULDER, O_DIAMOND, O_BRICKWALL, O_STEELWALL, O_START, O_EXIT,
  O_AMOEBA, O_BLOCK, O_EXPLOSIVE, O_FIREFLY, O_BUTTERFLY, O_SPACE, O_SPACE, O_SPACE
};

//
// -1  goto Intro
//  0  same Level again
//  1  next Level
//
static void decodeLevel(int gotoLevelFlag) {
  int x,y;
  int i;

  if (gotoLevelFlag<0) {
    curLevelOffs=0;
    lastLevel=curLevel;
    curLevel=0;    
  }
  if (gotoLevelFlag>0) {
    curLevelOffs=nextLevelOffs;
    curLevel++;
  }
  if (levelData[curLevelOffs]==0) {
    curLevelOffs=LEVEL1OFFS; // Wrap around after last level
    curLevel=1;
  }

  // Clear playarea
  memset(playArea, 0, sizeof(playArea));
  memset(moveArea, 0, sizeof(moveArea));
  int ld=curLevelOffs;

  levelX=levelData[ld++];
  levelY=levelData[ld++];
  diamondsNeeded=levelData[ld]*256+levelData[ld+1];
  ld+=2;

  for(i=0;i<4;i++) {
    colors[i]=levelData[ld]*65536+levelData[ld+1]*256+levelData[ld+2];
    colors[i+4]=colors[i];
    ld+=3;
  }

  int curCode;
  int codeLen=0;
  for(y=0;y<levelY;y++) {
    for(x=0;x<levelX;x++) {
      unsigned char c;

      if (y==0 || x==0 || y>(levelY-2) || x>(levelX-2)) {
        c=O_STEELWALL;
      } else {
        if (codeLen==0) {
          curCode=levelData[ld]&15;
          codeLen=levelData[ld++]>>4;
          if (codeLen==0) codeLen=levelData[ld++];
        }
        c=codeToObject[curCode];
        codeLen--;
      }
      playArea[y][x]=c;
    }
  }

  nextLevelOffs=ld;  
  cycle=0;
  keyboardEscape=0;
  diamondsCollected=0;
  rockfordX=0;
  rockfordY=0;
  screenX=64*32-SCREENX;
  screenY=64*32-SCREENY;
}

/*
  biSize; biWidth; biHeight; biPlanes; biBitCount;
  biCompression; biSizeImage; biXPelsPerMeter; biYPelsPerMeter; biClrUsed; biClrImportant;
*/
static const BITMAPINFO myBitmapInfo={
  sizeof(BITMAPINFOHEADER), BITMAPX, -BITMAPY, 1, 32,
  BI_RGB, 0, 0, 0, 0, 0
};


static void buttonPress(void) {
  keyboardPickup=1;
  if (curLevel==0) {
    // Start new game
    score=0;
    men=3;
    decodeLevel(1);
  } else if (curLevel>0 && cycle>2*STARTUPTIME && rockfortNotSeen>8) {
    men--;
    if (men==0) {
      decodeLevel(-1);
    } else {
      decodeLevel(0);
    }
  }
}


// 0x0001 0x0002 0x0004
// 0x0008 0x0010 0x0020
// 0x0040 0x0080 0x0100
// 0x0200 0x0400 0x0800
// 0x1000 0x2000 0x4000
static const short scoreStr[]={0x0000, 0x0000, 0x388E, 0x6270, 0x2B50, 0x1358, 0x63EA, 0x2080, -1};
static const short timeStr[]={0x0000, 0x0000, 0x2497, 0x2482, 0x5B7D, 0x63EA, 0x2080, -1};
//static const short diamondStr[]={0x0000, 0x3B6B, 0x2482, 0x5BEA, 0x5B7D, 0x2B6A, 0x5B58, 0x3B6B, 0x388E, 0x2080, -1};
//static const short diamondStr[]={0x3B6B, 0x2080, -1};
static const short levelStr[]={0x0000, 0x0000, 0x7249, 0x63EA, 0x2B40, 0x63EA, 0x7249, 0x2080, -1};
static const short menStr[]={0x0000, 0x0000, 0x5B7D, 0x63EA, 0x5B58, 0x2080, -1};
static LRESULT CALLBACK myWndProc(HWND hwnd,UINT uMsg,WPARAM wParam,LPARAM lParam) {
  switch(uMsg) {
  case WM_CREATE: {
      HDC hdc=GetDC(hwnd);
      playfieldDC=CreateCompatibleDC(hdc);
      playfieldBitmap=CreateDIBSection(playfieldDC, &myBitmapInfo, DIB_RGB_COLORS, (void **)&bitplanes, NULL, 0);
      SelectObject(playfieldDC, playfieldBitmap);
      ReleaseDC(hwnd, hdc);
      fillRndArray();

      decodeLevel(-1);

      SetTimer(hwnd, 1, 50, NULL);
    } break;
  case WM_CLOSE:
    PostQuitMessage(0);
    break;
/*  case WM_PAINT: {
      PAINTSTRUCT ps;
      HDC hdc;
      hdc=BeginPaint(hwnd, &ps);
      EndPaint(hwnd,&ps);
    } break;*/
  case WM_TIMER: {
      if ((cycle&3)==0) calculateArea();
      // Screen should follow rockford
/*
      if ((rockfordX*32-((3*SCREENX)>>2))>screenX && (SCREENX+screenX)<(levelX*32)) {
        screenX+=(cycle<STARTUPTIME)?64:8;
      }
      if ((rockfordX*32-(SCREENX>>2))<screenX || (SCREENX+screenX)>levelX*32) {
        screenX-=(cycle<STARTUPTIME)?64:8;
        if (screenX<0) screenX=0;
      }
      if ((rockfordY*32-((3*SCREENY)>>2))>screenY && (SCREENY+screenY)<(levelY*32+24)) {
        screenY+=(cycle<STARTUPTIME)?64:8;
      }
      if ((rockfordY*32-(SCREENY>>2))<screenY || (SCREENY+screenY)>levelY*32) {
        screenY-=(cycle<STARTUPTIME)?64:8;
        if (screenY<0) screenY=0;
      }
*/
      JOYINFO joyI;
      if (joyGetPos(JOYSTICKID1, &joyI)==JOYERR_NOERROR) {
        if (joyI.wButtons & (JOY_BUTTON1 | JOY_BUTTON2)) {
          buttonPress();
          useJoystick=1;
        }
        if (!(joyI.wButtons & (JOY_BUTTON1 | JOY_BUTTON2)) && useJoystick) {
          keyboardPickup=0;
        }
        if (useJoystick) {
          if (joyI.wXpos>49152) {
            keyboardX=1;
            keyboardY=0;
          } else if (joyI.wXpos<16384) {
            keyboardX=-1;
            keyboardY=0;
          } else if (joyI.wYpos>49152) {
            keyboardX=0;
            keyboardY=1;
          } else if (joyI.wYpos<16384) {
            keyboardX=0;
            keyboardY=-1;
          } else {
            keyboardX=0;
            keyboardY=0;
          }
        }
      }
      screenX+=((rockfordX*32-(screenX+(SCREENX/2)))+15)>>4;
      screenY+=((rockfordY*32-(screenY+(SCREENY/2)))+15)>>4;
//      (screenX*32+(rockfordX*32-SCREENX)+31)>>5;
//     screenY=(screenY*32+(rockfordY*32-SCREENY)+31)>>5;
      if (screenX+SCREENX>levelX*32) screenX=levelX*32-SCREENX;
      if (screenY+SCREENY>levelY*32) screenY=levelY*32-SCREENY;
      if (screenX<0) screenX=0;
      if (screenY<0) screenY=0;

      HDC hdc=GetDC(mainWindow);
        if ((cycle&3)==0) {
          clearInfoLine();
          drawInt( 0, diamondsCollected, 3, 10);
          drawChar( 3, 0x02A0);
          drawInt( 4, diamondsNeeded, 3, 10);
          drawStr( 7  , scoreStr);
          drawInt(15, score, 7, 10);
          if (curLevel) {
            if (cycle<STARTUPTIME) {
              drawStr(22, levelStr);
              drawInt(30, curLevel, 3, 10);
            } else {
              drawStr(22, timeStr);
              drawInt(29, 0, 4, 10);
            }
          } else {
            drawStr(22, levelStr);
            drawInt(30, lastLevel, 3, 10);
          }
          drawStr(33, menStr);
          drawInt(39, men, 1, 10);
          BitBlt(hdc, 0, 0, SCREENX, 24, playfieldDC, 0, 0, SRCCOPY);
        }
        displayPlayArea();
        BitBlt(hdc, 0, 24, SCREENX, SCREENY, playfieldDC, 32+(screenX&31), 32+(screenY&31), SRCCOPY);
        //if (startupmask<65535) startupmask+=256;
      ReleaseDC(mainWindow, hdc);
  //    InvalidateRect(hwnd, NULL, FALSE);
      cycle++;

      // Hack
      if (levelQuit) {
        levelQuit=false;
        decodeLevel(1);
      }
    } break;
  case WM_KEYDOWN:
    if (wParam==VK_UP) {
      keyboardX=0;
      keyboardY=-1;
      useJoystick=0;
    }
    if (wParam==VK_RIGHT) {
      keyboardX=1;
      keyboardY=0;
      useJoystick=0;
    }
    if (wParam==VK_DOWN) {
      keyboardX=0;
      keyboardY=1;
      useJoystick=0;
    }
    if (wParam==VK_LEFT) {
      keyboardX=-1;
      keyboardY=0;
      useJoystick=0;
    }
    if (wParam==VK_ESCAPE) {
      keyboardEscape=1;
    }
    if (wParam==VK_CONTROL || wParam==VK_SHIFT) {
      buttonPress();
      useJoystick=0;
    }
    if (wParam=='N' && rockfortNotSeen==0) {
      decodeLevel(1);
    }
    if ((wParam&0x40000000)==0) {
      // Only force first keypress.
      forcekeyboardX=keyboardX;
      forcekeyboardY=keyboardY;
    }
    break;
  case WM_KEYUP:
    if (VK_UP==wParam && keyboardY==-1) {
      keyboardY=0;
    }
    if (VK_RIGHT==wParam && keyboardX==1) {
      keyboardX=0;
    }
    if (VK_DOWN==wParam && keyboardY==1) {
      keyboardY=0;
    }
    if (VK_LEFT==wParam && keyboardX==-1) {
      keyboardX=0;
    }
    if (VK_CONTROL==wParam || VK_SHIFT==wParam) {
      keyboardPickup=0;
    }
    break;
  case WM_DESTROY:
    break;
  default:
    return DefWindowProc(hwnd,uMsg,wParam,lParam);
  }

  return 0;
}

static void InitMainWindow(int showMethod) {
  WNDCLASSEX myClass;

  myClass.cbSize=sizeof(myClass);
  myClass.style=CS_DBLCLKS; /* Enable double click processing */
  myClass.lpfnWndProc=myWndProc;
  myClass.cbClsExtra=0;
  myClass.cbWndExtra=0;
  myClass.hInstance=myInstance;
  myClass.hIcon=NULL;
  myClass.hCursor=LoadCursor(NULL, IDC_ARROW);
  myClass.hbrBackground=(HBRUSH)GetStockObject(GRAY_BRUSH);
  myClass.lpszMenuName=NULL;
  myClass.lpszClassName=className;
  myClass.hIconSm=NULL;
  mainClassAtom=RegisterClassEx(&myClass);

  int XSize=SCREENX+2*GetSystemMetrics(SM_CXFRAME);
  int YSize=SCREENY+GetSystemMetrics(SM_CYCAPTION)+2*GetSystemMetrics(SM_CYFRAME);

  // Create main window
  mainWindow=CreateWindowEx(0, className, windowTitle, WS_OVERLAPPED | WS_SYSMENU | WS_MINIMIZEBOX | WS_CLIPCHILDREN,
    (GetSystemMetrics(SM_CXFULLSCREEN)-XSize)/2, (GetSystemMetrics(SM_CYFULLSCREEN)-YSize)/2, XSize, YSize,
    NULL, NULL, myInstance, NULL);
  ShowWindow(mainWindow, showMethod);
  UpdateWindow(mainWindow);
}


#ifdef _DEBUG
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR szCmdLine, int iCmdShow) {
  myInstance=GetModuleHandle(NULL);
  InitMainWindow(iCmdShow);
//  InvalidateRect(mainWindow, NULL, TRUE);
  for(;;) {
    MSG myMsg;
    if (TRUE==GetMessage(&myMsg,NULL,0,0)) {
      TranslateMessage(&myMsg);
      DispatchMessage(&myMsg);
    } else {
      break;
    }
  }

  return 0;
}
#else
void WINAPI entry(void) {
  myInstance=GetModuleHandle(NULL);
  InitMainWindow(SW_SHOW);
  for(;;) {
    MSG myMsg;
    if (TRUE==GetMessage(&myMsg,NULL,0,0)) {
      TranslateMessage(&myMsg);
      DispatchMessage(&myMsg);
    } else {
      break;
    }
  }
}
#endif

